(******************************************************************************)
(* computer (accumulator machine)                                             *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Comp;

IMPORT FIO;

CONST
  debug = TRUE;
  out ::= FIO.StandardOutput;
  err ::= FIO.ErrorOutput;
  dbg ::= FIO.ErrorOutput;

TYPE
  TStr = ARRAY [ 0 .. 2 ] OF CHAR;

VAR
  acc: TValue;
  stk: ARRAY TIndex OF TValue; (* only vars on "stack" *)
  ops: ARRAY TOperation OF TStr;
  first: BOOLEAN;

PROCEDURE DbgH();
BEGIN
  FIO.WrStr( dbg, 'Comp.Debug: accumulator' );
  FIO.WrLn( dbg );
END DbgH;

PROCEDURE DbgS( s: ARRAY OF CHAR );
BEGIN
  IF first THEN DbgH(); first := FALSE END;
  FIO.WrStr( dbg, 'Comp.Debug: ' );
  FIO.WrInt( dbg, acc, 11 );
  FIO.WrStr( dbg, ':  ' );
  FIO.WrStr( dbg, s );
END DbgS;

PROCEDURE DbgX();
BEGIN
  FIO.WrLn( dbg );
END DbgX;

PROCEDURE DbgV( v: TValue );
BEGIN
  FIO.WrStr( dbg, '  ' );
  FIO.WrInt( dbg, v, 4 );
  FIO.WrLn( dbg );
END DbgV;

PROCEDURE DbgI( i: TIndex );
BEGIN
  FIO.WrStr( dbg, '  [' );
  FIO.WrCard( dbg, i, 2 );
  FIO.WrStr( dbg, ']' );
  FIO.WrLn( dbg );
END DbgI;


PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Comp: ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;


PROCEDURE Ldv( v: TValue );
BEGIN
  IF debug THEN DbgS( 'Lod' ); DbgV( v ) END;
  acc := v;
END Ldv;

PROCEDURE Ldi( i: TIndex );
BEGIN
  IF debug THEN DbgS( 'Lod' ); DbgI( i ) END;
  acc := stk[ i ];
END Ldi;

PROCEDURE Sto( i: TIndex );
BEGIN
  IF debug THEN DbgS( 'Sto' ); DbgI( i ) END;
  stk[ i ] := acc;
END Sto;

PROCEDURE Neg();
BEGIN
  IF debug THEN DbgS( 'Neg' ); DbgX() END;
  acc := -acc;
END Neg;

PROCEDURE Op0( o: TOperation; v: TValue );
BEGIN
  CASE o OF
  | oAdd: acc := acc + v;
  | oSub: acc := acc - v;
  | oMul: acc := acc * v;
  | oDiv: IF v = 0 THEN Err( 'Division by zero' ) END;
          acc := acc DIV v;
  ELSE
    Err( 'Bug: Unknown operation' );
  END;
END Op0;

PROCEDURE Opv( o: TOperation; v: TValue );
BEGIN
  IF debug THEN DbgS( ops[o] ); DbgV( v ) END;
  Op0( o, v );
END Opv;

PROCEDURE Opi( o: TOperation; i: TIndex );
BEGIN
  IF debug THEN DbgS( ops[o] ); DbgI( i ) END;
  Op0( o, stk[i] );
END Opi;

PROCEDURE Out();
BEGIN
  IF debug THEN DbgS( 'Out' ); DbgX() END;
  FIO.WrInt( out, acc, 0 ); FIO.WrLn( out );
END Out;


BEGIN
  ops[ oAdd ] := 'Add';
  ops[ oSub ] := 'Sub';
  ops[ oMul ] := 'Mul';
  ops[ oDiv ] := 'Div';
  first := TRUE;
END Comp.
